function make()
    !grep '^OPTFLAGS.*USE_MUMPS' ../../SuiteOPTconfig/Userconfig.mk > .temp
    use_mumps = fileread('.temp') ;
    !rm .temp
    use_harwell = 0 ;
    if ( length(use_mumps) > 5 )
        use_mumps = 1 ;
    else
        use_mumps = 0 ;
        !grep '^OPTFLAGS.*USE_HARWELL' ../../SuiteOPTconfig/Userconfig.mk >.temp
        use_harwell = fileread('.temp') ;
        !rm .temp
        if ( length(use_harwell) > 5 )
            use_harwell = 1 ;
        end
        if ( use_harwell == 1 )
            fprintf ('Currently, the Harwell routine MA57 is not\n') ;
            fprintf ('configured for use in SuiteOPT\n') ;
            error ('stop') ;
        end
    end
    if ( use_mumps || use_harwell )
        use_ss = 1 ;
    else
        use_ss = 0 ;
    end

    % if using mumps, then go to toplevel and complile mumps for MATLAB
    % (int64 & no DGEMMT)
    if ( use_mumps )
        cd ../../
        !make suitesparse % mumps needs compiled metis for examples
        !make mumpsmatlab
    end
    if ( use_ss )
        scotchpath = fileread('MATLAB/.scotch') ;
        !rm MATLAB/.scotch
        if ( length(scotchpath) > 2 )
            scotchpath = scotchpath(1:end-1) ;
            use_scotch = 1 ;
        else
            use_scotch = 0 ;
        end
        cd CGDESCENT/MATLAB ;
    end

    % location of SuiteSparse directory
    cd ../../SuiteSparseX
    SuiteSparsePath = strcat(pwd, '/') ;
    metis_path = strcat(SuiteSparsePath, 'metis-5.1.0/') ;
    CHOLMODpath = strcat(SuiteSparsePath, 'CHOLMOD/') ;
    config_path = strcat(SuiteSparsePath, 'SuiteSparse_config/') ;
    cd ..
    SuiteOPTpath = strcat(pwd, '/') ;
    cd CGDESCENT/MATLAB

    %% ---------------- Detect system running MATLAB ---------------- %%
    v = version ;
    try
        % ispc does not appear in MATLAB 5.3
        pc = ispc ;
        mac = ismac ;
    catch
        % if ispc fails, assume we are on a Windows PC if it's not unix
        pc = ~isunix ;
        mac = 0 ;
    end
    
    % Initialize flags variable to null 
    flags = '' ;
    is64 = ~isempty (strfind (computer, '64')) ;
    if (is64)
        % 64-bit MATLAB
        flags = [flags '-largeArrayDims'] ;
    end

    % MATLAB 8.3.0 now has a -silent option to keep 'mex' from burbling too much
    if (~verLessThan ('matlab', '8.3.0'))
        flags = ['-silent ' flags] ;
    end

    include = '' ;
    if (verLessThan ('matlab', '7.0'))
        % do not attempt to compile CHOLMOD with large file support
        include = [include ' -DNLARGEFILE'] ;
    elseif (~pc)
        % Linux/Unix require these flags for large file support
        include = [include ' -D_FILE_OFFSET_BITS=64 -D_LARGEFILE64_SOURCE'] ;
    end

    if (verLessThan ('matlab', '6.5'))
        % logical class does not exist in MATLAB 6.1 or earlier
        include = [include ' -DMATLAB6p1_OR_EARLIER'] ;
    end
    MATLABpath = strcat(SuiteSparsePath, 'CHOLMOD/MATLAB/') ;
    cholmod_matlab = { [MATLABpath 'cholmod_matlab'] } ;
    include = [include ' -I' MATLABpath] ;
    include = [include ' -I' SuiteSparsePath 'SuiteSparse_config/'] ;
    include = [include ' -I' CHOLMODpath 'Include'] ;

    %---------------------------------------------------------------------------
    % BLAS option
    %---------------------------------------------------------------------------

    % This is exceedingly ugly. The MATLAB mex command needs to be told where to
    % find the LAPACK and BLAS libraries, which is a real portability nightmare.

    if (pc)
        if (verLessThan ('matlab', '7.5'))
            lapack = 'libmwlapack.lib' ;
        elseif ( verLessThan ('matlab', '9.5') )
            lapack = 'libmwlapack.lib libmwblas.lib' ;
        else % version = 9.5 or newer
            lapack = '-lmwlapack -lmwblas' ;
        end
    else
        if (verLessThan ('matlab', '7.5'))
            lapack = '-lmwlapack' ;
        else
            lapack = '-lmwlapack -lmwblas' ;
        end
    end

    if (is64 && ~verLessThan ('matlab', '7.8'))
        % versions 7.8 and later on 64-bit platforms use a 64-bit BLAS
        fprintf ('with 64-bit BLAS\n') ;
        flags = [flags ' -DBLAS64'] ;
    end

    if (~(pc || mac))
        % for POSIX timing routine
        lapack = [lapack ' -lrt'] ;
    end

    if (pc)
        % Windows does not have drand48 and srand48, required by METIS.  Use
        % drand48 and srand48 in CHOLMOD/MATLAB/Windows/rand48.c instead.
        % Also provide Windows with an empty <strings.h> include file.
        obj_extension = '.obj' ;
        Windows_path = strcat(MATLABpath, 'Windows/') ;
        Windows_rand48 = strcat(MATLABpath, 'Windows/rand48') ;
        cholmod_matlab = [cholmod_matlab {Windows_rand48}] ;
        include = [include ' -I' Windows_path] ;
    else
        obj_extension = '.o' ;
    end

    % If a symmetric solver is used, then include the metis files if present
    metis_obj = '' ;
    metis_src = '' ;
    if ( use_ss )
        % Determine if METIS is available
        have_metis = exist (metis_path, 'dir') ;
        if (have_metis)
            fprintf ('Compiling METIS 5.1.0 for MATLAB Version %s\n', v) ;
            include = [include ' -I' metis_path 'include'] ;
            include = [include ' -I' metis_path 'GKlib'] ;
            include = [include ' -I' metis_path 'libmetis'] ;
        else
            fprintf ('Compiling without METIS for MATLAB Version %s\n', v) ;
            include = ['-DNPARTITION ' include] ;
        end

        if (have_metis)
            metis_src = {
                'GKlib/b64', ...
                'GKlib/blas', ...
                'GKlib/csr', ...
                'GKlib/error', ...
                'GKlib/evaluate', ...
                'GKlib/fkvkselect', ...
                'GKlib/fs', ...
                'GKlib/getopt', ...
                'GKlib/gkregex', ...
                'GKlib/graph', ...
                'GKlib/htable', ...
                'GKlib/io', ...
                'GKlib/itemsets', ...
                'GKlib/mcore', ...
                'GKlib/memory', ...
                'GKlib/omp', ...
                'GKlib/pdb', ...
                'GKlib/pqueue', ...
                'GKlib/random', ...
                'GKlib/rw', ...
                'GKlib/seq', ...
                'GKlib/sort', ...
                'GKlib/string', ...
                'GKlib/timers', ...
                'GKlib/tokenizer', ...
                'GKlib/util', ...
                'libmetis/auxapi', ...
                'libmetis/balance', ...
                'libmetis/bucketsort', ...
                'libmetis/checkgraph', ...
                'libmetis/coarsen', ...
                'libmetis/compress', ...
                'libmetis/contig', ...
                'libmetis/debug', ...
                'libmetis/fm', ...
                'libmetis/fortran', ...
                'libmetis/frename', ...
                'libmetis/gklib', ...
                'libmetis/graph', ...
                'libmetis/initpart', ...
                'libmetis/kmetis', ...
                'libmetis/kwayfm', ...
                'libmetis/kwayrefine', ...
                'libmetis/mcutil', ...
                'libmetis/mesh', ...
                'libmetis/meshpart', ...
                'libmetis/minconn', ...
                'libmetis/mincover', ...
                'libmetis/mmd', ...
                'libmetis/ometis', ...
                'libmetis/options', ...
                'libmetis/parmetis', ...
                'libmetis/pmetis', ...
                'libmetis/refine', ...
                'libmetis/separator', ...
                'libmetis/sfm', ...
                'libmetis/srefine', ...
                'libmetis/stat', ...
                'libmetis/timing', ...
                'libmetis/util', ...
                'libmetis/wspace'} ;
    
            for i = 1:length (metis_src)
                metis_src {i} = [metis_path metis_src{i}] ;
            end
        end
    end

    % compile the ordering routines from SuiteSparse if Harwell is used
    cholmod_obj = '' ;
    cholmod_src = '' ;
    ordering_src = '' ;
    config_src = '' ;
    if ( use_harwell )
        include = [include ' -I' SuiteSparsePath 'AMD/Include'] ;
        include = [include ' -I' SuiteSparsePath 'COLAMD/Include'] ;
        include = [include ' -I' SuiteSparsePath 'CCOLAMD/Include'] ;
        include = [include ' -I' SuiteSparsePath 'CAMD/Include'] ;
        include = [include ' -I' SuiteSparsePath 'CHOLMOD/Include'] ;
        include = [include ' -I' SuiteSparsePath 'CHOLMOD/MATLAB'] ;
    
        config_src = strcat(config_path, 'SuiteSparse_config') ;

        ordering_src = { ...
            'AMD/Source/amd_1', ...
            'AMD/Source/amd_2', ...
            'AMD/Source/amd_aat', ...
            'AMD/Source/amd_control', ...
            'AMD/Source/amd_defaults', ...
            'AMD/Source/amd_dump', ...
            'AMD/Source/amd_global', ...
            'AMD/Source/amd_info', ...
            'AMD/Source/amd_order', ...
            'AMD/Source/amd_postorder', ...
            'AMD/Source/amd_post_tree', ...
            'AMD/Source/amd_preprocess', ...
            'AMD/Source/amd_valid', ...
            'CAMD/Source/camd_1', ...
            'CAMD/Source/camd_2', ...
            'CAMD/Source/camd_aat', ...
            'CAMD/Source/camd_control', ...
            'CAMD/Source/camd_defaults', ...
            'CAMD/Source/camd_dump', ...
            'CAMD/Source/camd_global', ...
            'CAMD/Source/camd_info', ...
            'CAMD/Source/camd_order', ...
            'CAMD/Source/camd_postorder', ...
            'CAMD/Source/camd_preprocess', ...
            'CAMD/Source/camd_valid', ...
            'COLAMD/Source/colamd', ...
            'CCOLAMD/Source/ccolamd'} ;
    
        % Update ordering source to be located in SuiteSparse directory
        for k = 1:length(ordering_src)
            ordering_src {k} = strcat(SuiteSparsePath, ordering_src {k}) ;
        end
    
        cholmod_src = {
            'Core/cholmod_aat', ...
            'Core/cholmod_add', ...
            'Core/cholmod_band', ...
            'Core/cholmod_change_factor', ...
            'Core/cholmod_common', ...
            'Core/cholmod_complex', ...
            'Core/cholmod_copy', ...
            'Core/cholmod_dense', ...
            'Core/cholmod_error', ...
            'Core/cholmod_factor', ...
            'Core/cholmod_memory', ...
            'Core/cholmod_sparse', ...
            'Core/cholmod_transpose', ...
            'Core/cholmod_triplet', ...
            'Check/cholmod_check', ...
            'Check/cholmod_read', ...
            'Check/cholmod_write', ...
            'Cholesky/cholmod_amd', ...
            'Cholesky/cholmod_analyze', ...
            'Cholesky/cholmod_colamd', ...
            'Cholesky/cholmod_etree', ...
            'Cholesky/cholmod_factorize', ...
            'Cholesky/cholmod_postorder', ...
            'Cholesky/cholmod_rcond', ...
            'Cholesky/cholmod_resymbol', ...
            'Cholesky/cholmod_rowcolcounts', ...
            'Cholesky/cholmod_rowfac', ...
            'Cholesky/cholmod_solve', ...
            'Cholesky/cholmod_spsolve', ...
            'MatrixOps/cholmod_drop', ...
            'MatrixOps/cholmod_horzcat', ...
            'MatrixOps/cholmod_norm', ...
            'MatrixOps/cholmod_scale', ...
            'MatrixOps/cholmod_sdmult', ...
            'MatrixOps/cholmod_ssmult', ...
            'MatrixOps/cholmod_submatrix', ...
            'MatrixOps/cholmod_vertcat', ...
            'MatrixOps/cholmod_symmetry', ...
            'Modify/cholmod_rowadd', ...
            'Modify/cholmod_rowdel', ...
            'Modify/cholmod_updown', ...
            'Supernodal/cholmod_super_numeric', ...
            'Supernodal/cholmod_super_solve', ...
            'Supernodal/cholmod_super_symbolic', ...
            'Partition/cholmod_ccolamd', ...
            'Partition/cholmod_csymamd', ...
            'Partition/cholmod_camd', ...
            'Partition/cholmod_metis', ...
            'Partition/cholmod_nesdis'} ;
        % Update cholmod source to be located in CHOLMOD directory
        for k = 1:length(cholmod_src)
            cholmod_src{k} = [CHOLMODpath cholmod_src{k}] ;
        end
    end
    
    % compile each library source file
    source = [metis_src ordering_src config_src cholmod_src];
    if ( use_harwell )
        source = [source cholmod_matlab] ;
    end
    
    details = 0 ;
    kk = 0 ;
    
    %% ------------ Compile CHOLMOD objects from source code ------------ %%
    for f = source
        ff = f {1} ;
        if (isequal (ff, [metis_path 'GKlib/util']))
            % special case, since a file with the same name exists in libmetis
            copyfile ([ff '.c'], 'GKlib_util.c', 'f') ;
            ff = 'GKlib_util' ;
            o = 'GKlib_util' ;
        elseif (isequal (ff, [metis_path 'GKlib/graph']))
            % special case, since a file with the same name exists in libmetis
            copyfile ([ff '.c'], 'GKlib_graph.c', 'f') ;
            ff = 'GKlib_graph' ;
            o = 'GKlib_graph' ;
        else

            slash = strfind (ff, '/') ;
            if (isempty (slash))
                slash = 1 ;
            else
                slash = slash (end) + 1 ;
            end
            o = ff (slash:end) ;
        end
        cholmod_obj = [cholmod_obj  ' ' o obj_extension] ;
        fprintf('\n\n  Making object file %s.o\n\n', o) ;
        s = sprintf ('mex %s -DDLONG -O %s -c %s.c', ...
                      flags, include, ff) ;
        kk = do_cmd (s, kk, details) ;
    end

   %% ---------- Compile the CG_DESCENT object and mex files ---------- %%
    % Used for printing details when compiling
    details = 0 ;
    % Used during compilation of objects
    kk = 0 ;
    if ( use_mumps )
        flags = ['-DUSE_MUMPS ' flags] ;
    elseif ( use_harwell )
        flags = ['-DUSE_HARWELL ' flags] ;
    end

    %% ---------- Indicate include directories ---------- %%
    include = [include ' -I' SuiteOPTpath 'SuiteOPTconfig'] ;
    include = [include ' -I' SuiteOPTpath 'MATLAB/Include'] ;
    include = [include ' -I' SuiteOPTpath 'SSM/Include'] ;
    include = [include ' -I' SuiteOPTpath 'CGDESCENT/Include'] ;
    if ( use_mumps) 
        include = [include ' -I' SuiteOPTpath 'MUMPS/include'] ;
        include = [include ' -I' SuiteOPTpath 'MUMPS/libseq'] ;
        include = [include ' -I' SuiteOPTpath 'MUMPS/src'] ;
    elseif ( use_harwell )
        include = [include ' -I' SuiteOPTpath 'LBL/include'] ;
    end
    include = [include ' -I' SuiteSparsePath 'include'] ;

    % list of suiteopt source files to include in mex compilation
    cg_src = {
             'CGDESCENT/Source/cg_default', ...
             'CGDESCENT/Source/cg_descent', ...
             'CGDESCENT/Source/cg_print', ...
             'CGDESCENT/Source/cg_util', ...
             'SSM/Source/SSM', ...
             'SSM/Source/SSMdiagopt', ...
             'SSM/Source/SSMmult', ...
             'SSM/Source/SSMrefine', ...
             'SSM/Source/SSMtridiag', ...
             'SSM/Source/SSMprint', ...
             'SuiteOPTconfig/sopt', ...
             'MATLAB/Source/suiteopt_matlab', ...
             'MATLAB/Source/cg_descent_matlab'} ;
    for k = 1:length(cg_src)
        cg_src{k} = [SuiteOPTpath cg_src{k}] ;
    end

    %% ------------------- Required flags ------------------- %%
    % These flags are REQUIRED for the CGDESCENT mex function to compile
    % add '-DCGDESCENT_MATLAB' to indicate CGDESCENT
    flags = [flags ' -DCGDESCENT_MATLAB'] ;

    % add '-DSUITEOPT_MATLAB_OBJ' to indicate solver uses obj/grad/valgrad
    flags = [flags ' -DSUITEOPT_MATLAB_OBJ'] ;

    % add '-DSUITEOPT_MATLAB_CGHPROD' to indicate solver uses cghprod
    flags = [flags ' -DSUITEOPT_MATLAB_CGHPROD'] ;

    % add '-DSUITEOPT_MATLAB_HESSIAN' indicates solver uses objective Hessian */
    flags = [flags ' -DSUITEOPT_MATLAB_HESSIAN'] ;

    % add '-DSUITEOPT_MATLAB_QP' to indicate solver applies to QPs
    flags = [flags ' -DSUITEOPT_MATLAB_QP'] ;
    
    %% ------------------- Debugging flags ------------------- %%
    % add '-g' flag for debugging           
    % flags = [flags ' -g'] ; 
    
    % add '-v' flag for verbose mode           
    % flags = [flags ' -v'] ; 
    
    % add '-DNDEBUG' to turn off debugging for cg_descent source code
    flags = [flags ' -DNDEBUG'] ;

    % add '-DDEBUG_SUITEOPT_MEX' to turn on debugging for SuiteOPT mex funcs
    % flags = [flags ' -DDEBUG_SUITEOPT_MEX'] ;

    % add '-DDEBUG_CGDESCENT_MEX' to turn on debugging for cg_descent_mex func
    % flags = [flags ' -DDEBUG_CGDESCENT_MEX'] ;
    
    
    %% ---------------- Compile objects from source code ---------------- %%
    source = [cg_src];
    
    % Keep track of all object files compiled for later cleanup of directory
    cg_obj = '' ;
    
    % compile object files 
    for f = source 
        ff = f {1} ;
        slash = strfind (ff, '/') ;
        if (isempty (slash))
            slash = 1 ;
        else
            slash = slash (end) + 1 ;
        end
        o = ff (slash:end) ;
        cg_obj = [cg_obj ' ' o obj_extension] ;
        fprintf('\n\n  Making object file %s.o\n\n', o) ;
        s = sprintf('mex %s -DDLONG -O %s -c %s.c', flags, include, ff) ;
        kk = do_cmd (s, kk, details) ;
    end

 
    %% ---------------- Compile mex function(s) ---------------- %%
    % list of mex functions required in mex compilation of cg_descent
    cg_descent_mex_src = {'cg_descent'} ;

    % compile each mexFunction
    for f = cg_descent_mex_src
        fprintf('\n\n  Making mex function %s.mex\n\n', f{1}) ;
        if ( use_mumps )
            s = sprintf('mex LDFLAGS="$LDFLAGS -fopenmp" %s', flags) ;
            s = sprintf('%s -DDLONG %s %s_mex.c ', s, include, f{1}) ;
            MUMPSldlibs = ['-L' SuiteOPTpath 'MUMPS/lib -ldmumps '];
            MUMPSldlibs = [MUMPSldlibs '-lmumps_common -lmpiseq -lpord '] ;
            if ( use_scotch )
                MUMPSldlibs = [MUMPSldlibs '-L' scotchpath ' -lesmumps '] ;
                MUMPSldlibs = [MUMPSldlibs '-lscotch -lscotcherr'] ;
            end
            s = sprintf('%s %s %s ', s, MUMPSldlibs, cholmod_obj) ;
            s = sprintf('%s %s -lgfortran %s ', s, cg_obj, lapack) ;
            s = sprintf ('%s -lpthread -lm -output %s', s, f{1}) ;
        else
            s = sprintf('mex LDFLAGS="$LDFLAGS -fopenmp" %s ', flags) ;
            s = sprintf('%s -DDLONG %s %s_mex.c  %s %s %s ', ...
                        s, include, f{1}, cholmod_obj, cg_obj, lapack) ;
            s = sprintf('%s -lgfortran -lpthread -lm -output %s', s, f{1}) ;
        end
        kk = do_cmd (s, kk, details) ;
    end
 
    %% ---------------- Clean up and exit ---------------- %%
    if ( ~isempty (metis_obj) )
        s = ['delete ' metis_obj] ;
        do_cmd (s, kk, details) ;
    end
    if ( ~isempty (cholmod_obj) )
        s = ['delete ' cholmod_obj] ;
        do_cmd (s, kk, details) ;
    end
    if ( ~isempty (cg_obj) )
        s = ['delete ' cg_obj] ;
        do_cmd (s, kk, details) ;
    end
    if (exist ('GKlib_util.c', 'file'))
        delete ('GKlib_util.c') ;
    end
    if (exist ('GKlib_graph.c', 'file'))
        delete ('GKlib_graph.c') ;
    end

    fprintf ('\n  CGDESCENT successfully compiled\n') ;
    
    %---------------------------------------------------------------------------
    % Function used in installation routine
    function kk = do_cmd (s, kk, details)
         %DO_CMD: evaluate a command, and either print it or print a "."
        if (details)
            fprintf ('%s\n', s) ;
        else
            if (mod (kk, 60) == 0)
                fprintf ('\n') ;
            end
            kk = kk + 1 ;
            fprintf ('.') ;
        end
        eval (s) ; 
    end
end
