#---------------------------------------------------------------------------
# specify the C and Fortran compilers
#---------------------------------------------------------------------------
CC = gcc
FC = gfortran
#---------------------------------------------------------------------------
# set the optimization level
#---------------------------------------------------------------------------

# add -DUSE_MUMPS to utilize MUMPS in the Hessian-based implementation
# add -DNDEBUG to deactivate debugging in SuiteOPT
# add -DDLONG to specify LONG integers in SuiteOPT.
# The USE_ and DLONG flags must be on the same line as OPTFLAGS
OPTFLAGS = -O3 -DNDEBUG -DUSE_MUMPS

# For debugging:
#OPTFLAGS = -W -Wall -DUSE_MUMPS -pedantic -Wmissing-prototypes \
	-Wredundant-decls -Wnested-externs -Wdisabled-optimization \
	-fexceptions -Wno-parentheses -Wshadow -Wcast-align \
	-Winline -Wstrict-prototypes -Wno-unknown-pragmas -g -DNSUPER

#---------------------------------------------------------------------------
# specify the BLAS, LAPACK, and LDLIBS
#---------------------------------------------------------------------------

# Specify BLAS. For Intel's # mkl BLAS, something like the following may work.

# for ordinary ints i32
BLAS = -L/usr/lib/x86_64-linux-gnu -lmkl_intel_lp64 -lmkl_intel_thread \
	-lmkl_core -liomp5 -lpthread -lgfortran
LAPACK = -lmkl_lapack95_lp64
LIBBLAS = $(BLAS)

# for long ints i64 (compiling with -DDLONG added to OPTFLAGS)
#BLAS = -L/usr/lib/x86_64-linux-gnu -lmkl_intel_ilp64 -lmkl_intel_thread \
	-lmkl_core -liomp5 -lpthread -lgfortran
#LAPACK = -lmkl_lapack95_ilp64
#LIBBLAS = $(BLAS)

# Performance problems have been reported when using open blas with OpenMP
# threading in the supernodal nodal routines. When using open blas, it is
# recommended to turn off supernodal routines by adding -DNSUPER to OPTFLAGS:
#
#BLAS   = -lopenblas -lpthread -lgfortran
#LAPACK = -llapack
#OPTFLAGS += -DNSUPER

# For MUMPS, give location of scotch archive and the scotch include,
# when scotch is available. In Ubuntu, the locations below may work.
# If scotch not available, comment out the next 2 lines.
SCOTCH = /usr/lib/x86_64-linux-gnu/scotch-int64
ISCOTCH = -I/usr/include/scotch

CP ?= cp -f
# The sopt.h file contains a definition for an infinite floating
# point number based on the ANSI C99 and C90 standards. A different definition
# can be given here using a flag on OPTFLAGS. For example:

# OPTFLAGS += -Dinf=1e200
